/***************************************************************************
                         AbstractRenderCanvas.cpp
 ---------------------------------------------------------------------------
     begin                : Do Aug 14 2003
     copyright            : (C) 2002 by Holger Dammertz
     email                : holger.dammertz@informatik.uni-ulm.de
 ***************************************************************************/

#include "AbstractRenderCanvas.h"

#include "png.h"

AbstractRenderCanvas::AbstractRenderCanvas(int f_XRes, int f_YRes)
{
	m_XResolution = f_XRes;
	m_YResolution = f_YRes;
}

AbstractRenderCanvas::~AbstractRenderCanvas(void)
{
}

void AbstractRenderCanvas::SaveAsBMP(const char* f_pFilename)
{
	/* Create a 32-bit surface with the bytes of each pixel in R,G,B,A order,
	as expected by OpenGL for textures */
	SDL_Surface *surface;
	Uint32 rmask, gmask, bmask, amask;
	static Uint32 color;
	static unsigned char* pPosition;
	float r, g, b;

    /* SDL interprets each pixel as a 32-bit number, so our masks must depend
       on the endianness (byte order) of the machine */
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
    rmask = 0xff000000;
    gmask = 0x00ff0000;
    bmask = 0x0000ff00;
    amask = 0x000000ff;
#else
    rmask = 0x000000ff;
    gmask = 0x0000ff00;
    bmask = 0x00ff0000;
    amask = 0xff000000;
#endif

    surface = SDL_CreateRGBSurface(SDL_SWSURFACE, m_XResolution, m_YResolution, 32,
                                   rmask, gmask, bmask, amask);
    if(surface == NULL) {
        fprintf(stderr, "CreateRGBSurface failed: %s\n", SDL_GetError());
        exit(1);
    }

	for (int y = 0; y < m_YResolution; y++)
	{
		for (int x = 0; x < m_XResolution; x++)
		{
			GetPixelColor(x, y, r, g, b);
			color = SDL_MapRGB(surface->format, (Uint8)(r * 255.0f),(Uint8)(g * 255.0f),(Uint8)(b * 255.0f));

			*((Uint32 *)surface->pixels + y * surface->pitch/4 + x) = color;
		}
	}
	SDL_SaveBMP(surface, f_pFilename);
	printf("Saving %s.\n", f_pFilename);
	SDL_FreeSurface(surface);
}

void AbstractRenderCanvas::SaveAsPNG(const char* f_pFilename)
{
	png_bytep *row_pointers;
	png_structp png_ptr;
    png_infop info_ptr;
	float r, g, b;

	/* create file */
	FILE *fp = fopen(f_pFilename, "wb");
	if (!fp)
	{
		fprintf(stderr, "[write_png_file] File %s could not be opened for writing", f_pFilename);
		return;
	}

	/* initialize stuff */
	png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	
	if (!png_ptr)
	{
		fprintf(stderr, "[write_png_file] png_create_write_struct failed");
		return;
	}

	info_ptr = png_create_info_struct(png_ptr);
	if (!info_ptr) 
	{
		fprintf(stderr, "[write_png_file] png_create_write_struct failed");
		return;
	}

	if (setjmp(png_jmpbuf(png_ptr)))
	{
		fprintf(stderr, "[write_png_file] Error during init_io");
		return;
	}

	png_init_io(png_ptr, fp);

	png_set_IHDR(png_ptr, info_ptr, m_XResolution, m_YResolution, 8, PNG_COLOR_TYPE_RGB,
      PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

	png_set_text(png_ptr, info_ptr, NULL, 0);

	png_write_info(png_ptr, info_ptr);

	/* write bytes */
	if (setjmp(png_jmpbuf(png_ptr)))
	{
		fprintf(stderr, "[write_png_file] Error during writing bytes");
		return;
	}

	// copy the data:
	row_pointers = (png_bytep*) malloc(sizeof(png_bytep) * m_YResolution);
	for (int y=0; y<m_YResolution; y++)
		row_pointers[y] = (png_byte*) malloc(info_ptr->rowbytes);

	for (int y = 0; y < m_YResolution; y++)
		for (int x = 0; x < m_XResolution; x++)
		{
			GetPixelColor(x, y, r, g, b);
			row_pointers[y][x*3+0] = (unsigned char)(r * 255.0f);
			row_pointers[y][x*3+1] = (unsigned char)(g * 255.0f);
			row_pointers[y][x*3+2] = (unsigned char)(b * 255.0f);
		}

	png_write_image(png_ptr, row_pointers);

	/* end write */
	if (setjmp(png_jmpbuf(png_ptr)))
	{
		fprintf(stderr, "[write_png_file] Error during end of write");
		return;
	}

	png_write_end(png_ptr, NULL);
	fclose(fp);
	printf("Saving %s.\n", f_pFilename);
}
