/***************************************************************************
                         SDLRenderCanvas.cpp
 ---------------------------------------------------------------------------
     begin                : Do Aug 14 2003
     copyright            : (C) 2002 by Holger Dammertz
     email                : holger.dammertz@informatik.uni-ulm.de
 ***************************************************************************/

#include "SDLRenderCanvas.h"

#include <stdlib.h>
#include <memory.h>

SDLRenderCanvas::SDLRenderCanvas(int f_XRes, int f_YRes, bool f_FullScreen, bool f_DoubleBuffer) : AbstractRenderCanvas(f_XRes, f_YRes)
{
	m_pScreen = NULL;
	m_pVideoInfo = NULL;
	m_fpQuitFunction = NULL;


	m_DoubleBuffered = f_DoubleBuffer;
	m_FullScreen = f_FullScreen;

	m_FrameCountInterval = 5000;
	m_ElapsedTicks = 0;
}


SDLRenderCanvas::~SDLRenderCanvas(void)
{
}



bool SDLRenderCanvas::Initialize()
{
	if (m_pScreen != NULL)
	{
		fprintf(stderr, "SDLRenderCanvas::Initialize(): There is already a screen "
			"initialized.\n");
		//return false;
	}
	if (SDL_Init(SDL_INIT_VIDEO) < 0)
	{
		fprintf(stderr, "SDLRenderCanvas::Initialize(): Video initialization failed: %s\n", 
			     SDL_GetError());
		return false;
	}

	SetVideoFlags(true);

	/* now we try to become a surface for rendering */
	m_pScreen = SDL_SetVideoMode(m_XResolution, m_YResolution, 32, m_VideoFlags);

	/* Verify that there is a surface */
	if (!m_pScreen)
	{
		fprintf(stderr, "SDLRenderCanvas::Initialize(): Video mode set failed: %s\n",
			SDL_GetError( ));
		return false;
	}

	fprintf(stderr, "Screen set to %d x %d x %d\n",  m_pScreen->w,
	       m_pScreen->h, m_pScreen->format->BitsPerPixel);

    return true;
}



void SDLRenderCanvas::SetQuitFunction(void (*f_fpQuitFunction)(int))
{
	m_fpQuitFunction = f_fpQuitFunction;
}



void SDLRenderCanvas::HandleEvents()
{
	bool consumed = false;

	while (SDL_PollEvent(&m_Event))
	{
		for (SDLEventConsumerVector::iterator i = m_EventConsumer.begin();
			i != m_EventConsumer.end(); i++)
		{
			consumed = (*i)->Consume(m_Event);
			// At the moment do not handle consumed events any further
			if (consumed) return;
		}
		
		switch(m_Event.type)
		{
		case SDL_QUIT:
			SDL_Quit();
			if (m_fpQuitFunction == NULL)
			{
				fprintf(stderr, "SDLRenderCanvas::HandleEvents(): There was no quit"
					" function set! Terminating anyway.\n");
				exit(0);
			}
			m_fpQuitFunction(0);
		case SDL_KEYDOWN:
			if (m_Event.key.keysym.sym == SDLK_ESCAPE)
			{
				SDL_Quit();
				if (m_fpQuitFunction == NULL)
				{
					fprintf(stderr, "SDLRenderCanvas::HandleEvents(): There was no quit"
						" function set! Terminating anyway.\n");
					exit(0);
				}
				m_fpQuitFunction(0);
			}
		default:
			break;
		}
	}
}


void SDLRenderCanvas::EnterRenderMode()
{
	if (SDL_MUSTLOCK(m_pScreen)) 
	{
		while(SDL_LockSurface(m_pScreen) < 0);
	}
}

void SDLRenderCanvas::LeaveRenderMode()
{
	// some static variables for fps counting
	static int oldTicks = 0;
	static int t0;
	static int frames;


	if (SDL_MUSTLOCK(m_pScreen)) SDL_UnlockSurface(m_pScreen);
	
	if (m_DoubleBuffered) SDL_Flip(m_pScreen);


	/* gather frames per second */
	frames++;
	{
		int t = SDL_GetTicks();
		if (t - t0 >= m_FrameCountInterval) 
		{
			double seconds = (t - t0) / 1000.0;
			double fps = frames / seconds;
			//sprintf(windowTitle, "%d frames in %g seconds = %g FPS", frames, seconds, fps);
			sprintf(m_pWindowTitle, "%g fps", fps);
			// Call this only, when you have an SDL window openend!
			SDL_WM_SetCaption(m_pWindowTitle, NULL);
			t0 = t;
			frames = 0;
		}                                
	}

	// Calculate the new DeltaT
	m_DeltaT = (float)(SDL_GetTicks()-m_ElapsedTicks)/1000.0f;
	m_ElapsedTicks = SDL_GetTicks();
}


void SDLRenderCanvas::LeaveRenderMode(int f_X, int f_Y, int f_W, int f_H)
{
	#ifdef DEBUG
		if (m_DoubleBuffered) logError("SDLRenderCanvas::LeaveRenderMode"
			"(int f_X, int f_Y, int f_W, int f_H): called in double buffered mode!");
	#endif
	
	SDL_UpdateRect(m_pScreen, f_X, f_Y, f_W, f_H);	
}


void SDLRenderCanvas::DrawPixel(int f_X, int f_Y, float f_R, float f_G, float f_B)
{
	static Uint32 color;
	static unsigned char* pPosition;

	color = SDL_MapRGB(m_pScreen->format, (Uint8)(f_R * 255), 
		               (Uint8)(f_G * 255), (Uint8)(f_B * 255));

	*((Uint32 *)m_pScreen->pixels + f_Y * m_pScreen->pitch/4 + f_X) = color;
}

void SDLRenderCanvas::DrawSafePixel(int f_X, int f_Y, float f_R, float f_G, float f_B)
{
	if ((f_X >= 0) && (f_X < m_XResolution) && (f_Y >= 0) && (f_Y < m_YResolution))
	{
		static Uint32 color;
		static unsigned char* pPosition;

		color = SDL_MapRGB(m_pScreen->format, (Uint8)(f_R * 255), 
			(Uint8)(f_G * 255), (Uint8)(f_B * 255));

		*((Uint32 *)m_pScreen->pixels + f_Y * m_pScreen->pitch/4 + f_X) = color;
	}
}


void SDLRenderCanvas::DrawPixelUchar(int f_X, int f_Y, unsigned char f_R, unsigned char  f_G, unsigned char  f_B)
{
	static Uint32 color;
	static unsigned char* pPosition;

	// !!OPT: This is slow
	color = SDL_MapRGB(m_pScreen->format, f_R, f_G, f_B);

	*((Uint32 *)m_pScreen->pixels + f_Y * m_pScreen->pitch/4 + f_X) = color;
}

void SDLRenderCanvas::GetPixelColor(int f_X, int f_Y, float& f_R, float& f_G, float& f_B)
{
	Uint8 red, green, blue, alpha;
	SDL_PixelFormat *fmt;
	int bpp;
	Uint32 temp, pixel;

	fmt=m_pScreen->format;

	// Lock the m_pScreen
	SDL_LockSurface(m_pScreen);

	bpp = m_pScreen->format->BytesPerPixel;
	
	if (bpp != 4) fprintf(stderr, "\nWARNING: 32 bits/pixel assumed!");

    // Here p is the address to the pixel we want to retrieve
    Uint8 *p = (Uint8 *)m_pScreen->pixels + f_Y * m_pScreen->pitch + f_X * bpp;
	pixel = *(Uint32 *)p;
	SDL_UnlockSurface(m_pScreen);

	/* Get Red component */
	temp=pixel&fmt->Rmask; /* Isolate red component */
	temp=temp>>fmt->Rshift;/* Shift it down to 8-bit */
	temp=temp<<fmt->Rloss; /* Expand to a full 8-bit number */
	red=(Uint8)temp;

	/* Get Green component */
	temp=pixel&fmt->Gmask; /* Isolate green component */
	temp=temp>>fmt->Gshift;/* Shift it down to 8-bit */
	temp=temp<<fmt->Gloss; /* Expand to a full 8-bit number */
	green=(Uint8)temp;

	/* Get Blue component */
	temp=pixel&fmt->Bmask; /* Isolate blue component */
	temp=temp>>fmt->Bshift;/* Shift it down to 8-bit */
	temp=temp<<fmt->Bloss; /* Expand to a full 8-bit number */
	blue=(Uint8)temp;

	/* Get Alpha component */
	temp=pixel&fmt->Amask; /* Isolate alpha component */
	temp=temp>>fmt->Ashift;/* Shift it down to 8-bit */
	temp=temp<<fmt->Aloss; /* Expand to a full 8-bit number */
	alpha=(Uint8)temp;

	f_R = red/255.0f, f_G = green/255.0f, f_B = blue/255.0f;	
}


void SDLRenderCanvas::AddEventConsumer(SDLEventConsumer* f_pConsumer)
{
	m_EventConsumer.push_back(f_pConsumer);
}

void SDLRenderCanvas::SaveToFile(const char* f_pFilename)
{
	SDL_SaveBMP(m_pScreen, f_pFilename);
	printf("Saving %s.\n", f_pFilename);
}

void SDLRenderCanvas::clearCanvas(float f_R, float f_G, float f_B)
{
	for (int y = 0; y < m_YResolution; y++)
	{
		for (int x = 0;  x < m_XResolution; x++)
		{
			DrawPixel(x, y, f_R, f_G, f_B);
		}
	}
}

/***************************************************************************
 *    Protected Methods
 ***************************************************************************/


void SDLRenderCanvas::SetVideoFlags(bool f_PrintInfo)
{
	m_pVideoInfo = SDL_GetVideoInfo();
	if (!m_pVideoInfo)
	{
		fprintf(stderr, "SDLRenderCanvas::SetVideoFlags(bool f_PrintInfo): Could"
			" not get VideoInfo from SDL: %s\n", SDL_GetError());
		return;
	}

	m_VideoFlags = SDL_HWPALETTE;  // Store the palette in hardware
	
	if (m_DoubleBuffered) m_VideoFlags |= SDL_DOUBLEBUF;    // Enable double buffering
	if (m_FullScreen) m_VideoFlags |= SDL_FULLSCREEN;       // Enable fullscreen
	if (m_pVideoInfo->blit_hw) m_VideoFlags |= SDL_HWACCEL; // This checks if hardware blits can be done 

	/* This checks to see if surfaces can be stored in memory */
	if (f_PrintInfo)
	{
		//logMessage("  Resolution: %dx%d", m_XResolution, m_YResolution);
		if (m_pVideoInfo->hw_available)
		{
			m_VideoFlags |= SDL_HWSURFACE;
			printf("  Surface is hardware.\n");
		}
		else
		{
			m_VideoFlags |= SDL_SWSURFACE;
			printf("  Surface is software.\n");
		}
		if (m_pVideoInfo->blit_hw)
		{
			printf("  Hardware accelerated blit enabled.\n");
		}
		else
		{
			printf("  Hardware accelerated blit disabled.\n");
		}
	}
}
