/***************************************************************************
                         SDLRenderCanvas.h
 ---------------------------------------------------------------------------
     begin                : Do Aug 14 2003
     copyright            : (C) 2002 by Holger Dammertz, Sabrina Skibak
     email                : holger.dammertz@informatik.uni-ulm.de
							sabrina.skibak@informatik.uni-ulm.de
 ***************************************************************************/

#ifndef SDLRENDERCANVAS_H
#define SDLRENDERCANVAS_H

#include <windows.h>

#include "SDL/SDL.h"
#include "AbstractRenderCanvas.h"

#include <vector>

class SDLEventConsumer
{
public:
	virtual bool Consume(SDL_Event& m_rEvent) = 0;
};

typedef std::vector<SDLEventConsumer*> SDLEventConsumerVector;

/**
 * This Implementation of AbstractRenderCanvas features a double buffered 
 * window that is created on initialization of the class. It uses the
 * SDL to create the windows.
 *
 * @TODO:
 *  - Make this class some kind of singelton because you can only open
 *    one SDL window at a time.
 */
class SDLRenderCanvas : public AbstractRenderCanvas
{
	/** A pointer to our screen */
	SDL_Surface* m_pScreen;
	/** stores some information */
	const SDL_VideoInfo* m_pVideoInfo;
	/** The video flags that define the propperties of the screen */
	int m_VideoFlags;
	/** to store if we have a double bufferd screen or not */
	bool m_DoubleBuffered;

	bool m_FullScreen;

	/** This is used for event handling (at the moment only checking 
	    for an exit action)	*/
	SDL_Event m_Event;
	/** This is a pointer to the function that is called on a quit event */
	void (*m_fpQuitFunction)(int);

	SDLEventConsumerVector m_EventConsumer;

	/** the window title (at the moment only used by the fps display */
	char m_pWindowTitle[255];
	/** the time in milliseconds that has to elapse before a new fps value
	    is evaluated */
	int m_FrameCountInterval;
	/** needed for deltaT calculation */
	int m_ElapsedTicks;

  public:
	SDLRenderCanvas(int f_XRes, int f_YRes, bool f_FullScreen, bool f_DoubleBuffer);
	virtual ~SDLRenderCanvas(void);

	virtual bool Initialize();

	virtual void SetQuitFunction(void (*f_fpQuitFunction)(int));

	/**
	 * This method checks for quit events (ESC or the close button of the
	 * window) and then calls the quit-function.
	 */
	virtual void HandleEvents();

	/**
	 * Make sure you call this method once before performing a DrawPixel call.
	 * It locks the surface so you can draw to it. When the drawing has finished,
	 * call a LeaveRenderMode();
	 */
	virtual void EnterRenderMode();

	/**
	 * This method unlocks the screen surface if needed.
	 * If double bufferig is enabled, it performas a buffer swap so the drawn 
	 * image can be seen. If not, you have to make sure your on your own, that
	 * the correct screen space is updated.
	 */
	virtual void LeaveRenderMode();

	/**
	 * This method unlocks the screen surface if needed and updates only the given
	 * rectangular region.
	 * (0, 0, 0, 0) Updates the whole screen
	 * @remarks
	 *   call this method only in non doublebufferd mode.
	 */
	virtual void LeaveRenderMode(int f_X, int f_Y, int f_W, int f_H);

	/**
	 * Expects the color values to be between 0 and 1 and scales them to 
	 * 0..255. At the moment this method uses a really slow color conversion
	 * routine...
	 */
	virtual void DrawPixel(int f_X, int f_Y, float f_R, float f_G, float f_B);

	//does some boundary checking
	virtual void DrawSafePixel(int f_X, int f_Y, float f_R, float f_G, float f_B);

	virtual void DrawPixelUchar(int f_X, int f_Y, unsigned char f_R, unsigned char  f_G, unsigned char  f_B);

	//Assumption: 32 bits/pixel!!!!
	virtual void GetPixelColor(int f_X, int f_Y, float& f_R, float& f_G, float& f_B);

	virtual void AddEventConsumer(SDLEventConsumer* f_pConsumer);

	// saves canvas as bmp
	virtual void SaveToFile(const char* f_pFilename);

	virtual void clearCanvas(float f_R, float f_G, float f_B);


  protected:
	virtual void SetVideoFlags(bool f_PrintInfo);
};

#endif