#ifndef RTMATH_H
#define RTMATH_H

#include <windows.h>
#include <math.h>

///////////////////////////////////////////////////////////////////////////////

static const float DIV_EPSILON = (1e-10f); // Avoid numerically instable divisions

///////////////////////////////////////////////////////////////////////////////

class Vector
{
public:
	float  dx, dy, dz;

	inline Vector();
	inline Vector(const float, const float, const float);

	inline Vector operator - () const;                   // Change Sign
	inline Vector operator + (const Vector&) const;
	inline Vector& operator += (const Vector&);
	inline Vector operator - (const Vector&) const;
	inline Vector& operator -= (const Vector&);

	inline Vector operator | (const Vector&) const;      // Cross product
	inline float operator * (const Vector&) const;       // Dot product

	inline Vector& operator *= (const float);
	inline Vector operator * (const float) const;
	inline friend Vector operator * (const float, const Vector&);
	inline Vector& operator /= (const float);
	inline Vector operator / (const float) const;

	inline friend float Length(const Vector&);
	inline friend float Normalize(Vector&);

	inline friend int GetMaxIndex(Vector&);

	inline float& operator[](const int i);
	inline float operator[](const int i) const;
};

///////////////////////////////////////////////////////////////////////////////

class Point
{
public:
	float  x, y, z;

	inline Point();
	inline Point(const float, const float, const float);

	inline Point operator + (const Vector&) const;
	inline Point& operator += (const Vector&);
	inline Point operator - (const Vector&) const;
	inline Point& operator -= (const Vector&);
	inline Vector operator - (const Point&) const;

	inline float& operator[](const int i);
	inline float operator[](const int i) const;
};

///////////////////////////////////////////////////////////////////////////////


inline Vector::Vector()
{
}

inline Vector::Vector(const float dxIn, const float dyIn, const float dzIn)
{
	dx = dxIn;
	dy = dyIn;
	dz = dzIn;
}

inline Vector Vector::operator - () const
{
	return Vector(- dx, - dy, - dz);
}

inline Vector Vector::operator + (const Vector& v) const
{
	return Vector(dx + v.dx, dy + v.dy, dz + v.dz);
}

inline Vector& Vector::operator += (const Vector& v)
{
	dx += v.dx;
	dy += v.dy;
	dz += v.dz;

	return *this;
}

inline Vector Vector::operator - (const Vector& v) const
{
	return Vector(dx - v.dx, dy - v.dy, dz - v.dz);
}

inline Vector& Vector::operator -= (const Vector& v)
{
	dx -= v.dx;
	dy -= v.dy;
	dz -= v.dz;

	return *this;
}

inline Vector Vector::operator | (const Vector& v) const
{
	return Vector(dy * v.dz - dz * v.dy, dz * v.dx - dx * v.dz, dx * v.dy - dy * v.dx);
}

inline float Vector::operator * (const Vector& v) const
{
	return (dx * v.dx + dy * v.dy + dz * v.dz);
}

inline Vector& Vector::operator *= (const float f)
{
	dx *= f;
	dy *= f;
	dz *= f;

	return *this;
}

inline Vector Vector::operator * (const float f) const
{
	return Vector(dx * f, dy * f, dz * f);
}

inline Vector operator * (const float f, const Vector& v)
{
	return v * f;
}

inline Vector& Vector::operator /= (const float f)
{
	*this *= 1.0f / f;

	return *this;
}

inline Vector Vector::operator / (const float f) const
{
	float F = 1.0f / f;

	return Vector(dx * F, dy * F, dz * F);
}

inline float Length(const Vector& v)
{
	return sqrtf(v * v);
}

inline float Normalize(Vector& v)
{
	float l = Length(v);

	if(l > DIV_EPSILON)
		v /= l;

	return l;
}

inline int GetMaxIndex(Vector& v)
{
	if (v.dx > v.dy)
	{
		if (v.dx > v.dz) return 0;
		else return 2;
	}
	else if (v.dy > v.dz) return 1;
	else return 2;
}


inline float& Vector::operator[](const int i)
{
	return (&dx)[i];
}

inline float Vector::operator[](const int i) const
{
	return (&dx)[i];
}

/////////////////////////////////////////////////////////////////////////////

inline Point::Point()
{
}

inline Point::Point(const float xIn, const float yIn, const float zIn)
{
	x = xIn;
	y = yIn;
	z = zIn;
}
inline Point Point::operator + (const Vector& v) const
{
	return Point(x + v.dx, y + v.dy, z + v.dz);
}

inline Point& Point::operator += (const Vector& v)
{
	x += v.dx;
	y += v.dy;
	z += v.dz;

	return *this;
}

inline Point Point::operator - (const Vector& v) const
{
	return Point(x - v.dx, y - v.dy, z - v.dz);
}

inline Point& Point::operator -= (const Vector& v)
{
	x -= v.dx;
	y -= v.dy;
	z -= v.dz;

	return *this;
}

inline Vector Point::operator - (const Point& P) const
{
	return Vector(x - P.x, y - P.y, z - P.z);
}

inline float& Point::operator[](const int i)
{
	return (&x)[i];
}

inline float Point::operator[](const int i) const
{
	return (&x)[i];
}

#endif